import pyomo.environ as pyo


class SimpleModel:
    """
    A simple resource allocation model with 3 products and 2 resources.
    """

    def __init__(self):
        cake_types = ['chocolate', 'vanilla', 'matcha']
        ingredients = ['eggs', 'flour']

        prices = {'chocolate': 3, 'vanilla': 4, 'matcha': 5}

        available_ingredients = {'eggs': 32, 'flour': 48}

        recipes = {
            ('eggs', 'chocolate'): 4,
            ('eggs', 'vanilla'): 2,
            ('eggs', 'matcha'): 3,
            ('flour', 'chocolate'): 4,
            ('flour', 'vanilla'): 6,
            ('flour', 'matcha'): 5,
        }

        self.model = pyo.ConcreteModel()

        #
        # Sets
        #
        self.model.K = pyo.Set(initialize=cake_types)
        self.model.I = pyo.Set(initialize=ingredients)

        #
        # Parameters
        #
        self.model.p = pyo.Param(self.model.K, initialize=prices)
        self.model.b = pyo.Param(self.model.I, initialize=available_ingredients)
        self.model.a = pyo.Param(self.model.I, self.model.K, initialize=recipes)

        #
        # Variables
        #
        self.model.x = pyo.Var(self.model.K, domain=pyo.Integers, bounds=(0, 100))

        #
        # Constraints
        #
        self.model.total_cakes = pyo.Expression(
            expr=pyo.quicksum(self.model.x[k] for k in self.model.K)
        )

        def ingredients_available_rule(model, i):
            return (
                pyo.quicksum(model.a[i, k] * model.x[k] for k in model.K) <= model.b[i]
            )

        self.model.ingredients_available_con = pyo.Constraint(
            self.model.I, rule=ingredients_available_rule
        )

        def total_cakes_rule(model):
            return model.total_cakes >= 4

        self.model.total_cakes_con = pyo.Constraint(rule=total_cakes_rule)

        #
        # Objective
        #
        def obj_rule(model):
            return 150 + pyo.quicksum(model.p[k] * model.x[k] for k in model.K)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class MaxObjModel:
    """
    A model with a maximization objective.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(0, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return pyo.quicksum(model.x[i] for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class MinObjModel:
    """
    A model with a minimization objective.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(0, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return pyo.quicksum(model.x[i] for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.minimize)


class RealVarsModel:
    """
    A model with real variables.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Reals, bounds=(0, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return pyo.quicksum(model.x[i] for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class NoLbVarsModel:
    """
    A model with variables unbounded from below.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(None, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return pyo.quicksum(model.x[i] for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class NoUbVarsModel:
    """
    A model with variables unbounded from above.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(0, None))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return pyo.quicksum(model.x[i] for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class QuadConModel:
    """
    A model with a quadratic constraint.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(0, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] ** 3 for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return pyo.quicksum(model.x[i] for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class NonlinearConModel:
    """
    A model with a nonlinear (but not quadratic) constraint.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(0, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * pyo.sin(model.x[i]) for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return pyo.quicksum(model.x[i] for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class QuadObjModel:
    """
    A model with a quadratic objective.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(0, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return pyo.quicksum(model.x[i] ** 2 for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class NonlinearObjModel:
    """
    A model with a nonlinear (but not quadratic) objective.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(0, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return pyo.quicksum(pyo.cos(model.x[i]) for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class InfeasibleModel:
    """
    An infeasible model.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(10, 100))

        def infeasible_con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 20

        self.model.infeasible_con = pyo.Constraint(rule=infeasible_con_rule)

        def obj_rule(model):
            return pyo.quicksum(model.x[i] for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class InactiveConModel:
    """
    A model with an inactive constraint.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(0, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 20

        self.model.con = pyo.Constraint(rule=con_rule)

        def inactive_con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) >= 100

        self.model.inactive_con = pyo.Constraint(rule=inactive_con_rule)
        self.model.inactive_con.deactivate()

        def obj_rule(model):
            return pyo.quicksum(model.x[i] for i in model.I)

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)


class ConstantObjModel:
    """
    A model with a constant objective.
    """

    def __init__(self):
        self.model = pyo.ConcreteModel()

        self.model.I = pyo.Set(initialize=[1, 2, 3])
        self.model.w = pyo.Param(self.model.I, initialize={1: 10, 2: 20, 3: 30})
        self.model.x = pyo.Var(self.model.I, domain=pyo.Integers, bounds=(0, 100))

        def con_rule(model):
            return pyo.quicksum(model.w[i] * model.x[i] for i in model.I) <= 10

        self.model.con = pyo.Constraint(rule=con_rule)

        def obj_rule(model):
            return 500 + 0 * model.x[1]

        self.model.obj = pyo.Objective(rule=obj_rule, sense=pyo.maximize)



import pytest
import pyomo.environ as pyo
from pyomo.contrib.solver.common.results import SolutionStatus, TerminationCondition
from pyomo.contrib.solver.common.util import (
    NoFeasibleSolutionError,
    NoOptimalSolutionError,
)
from pyomo_cpsat import Cpsat, IncompatibleModelError
from model import (
    SimpleModel,
    MinObjModel,
    MaxObjModel,
    RealVarsModel,
    NoLbVarsModel,
    NoUbVarsModel,
    QuadConModel,
    NonlinearConModel,
    QuadObjModel,
    NonlinearObjModel,
    InfeasibleModel,
    InactiveConModel,
    ConstantObjModel,
)


solver = Cpsat()


## Start tests
def test_max_objective():
    maxobj = MaxObjModel()
    solver.solve(maxobj.model)
    assert solver._solver_model.proto.objective.scaling_factor < 0


def test_min_objective():
    minobj = MinObjModel()
    solver.solve(minobj.model)
    assert solver._solver_model.proto.objective.scaling_factor >= 0


def test_solver_options():
    simple = SimpleModel()
    solver.solve(simple.model, solver_options={'num_full_subsolvers': 2})
    assert solver._solver_solver.parameters.num_full_subsolvers == 2


def test_solver_options_repeating():
    simple = SimpleModel()
    solver.solve(
        simple.model,
        solver_options={'subsolvers': ['pseudo_costs', 'probing']},
    )
    assert solver._solver_solver.parameters.subsolvers == ['pseudo_costs', 'probing']


def test_pyomo_equivalent_keys_threads():
    with pytest.raises(KeyError):
        simple = SimpleModel()
        solver.solve(
            simple.model,
            threads=1,
            solver_options={
                'num_workers': 1,
            },
        )


def test_pyomo_equivalent_keys_time_limit():
    with pytest.raises(KeyError):
        simple = SimpleModel()
        solver.solve(
            simple.model,
            time_limit=100,
            solver_options={
                'max_time_in_seconds': 100,
            },
        )


def test_pyomo_equivalent_keys_rel_gap():
    with pytest.raises(KeyError):
        simple = SimpleModel()
        solver.solve(
            simple.model,
            rel_gap=0.0,
            solver_options={
                'relative_gap_limit': 0.0,
            },
        )


def test_pyomo_equivalent_keys_abs_gap():
    with pytest.raises(KeyError):
        simple = SimpleModel()
        solver.solve(
            simple.model,
            abs_gap=1e-4,
            solver_options={
                'absolute_gap_limit': 1e-4,
            },
        )


def test_realvars():
    with pytest.raises(IncompatibleModelError):
        realvars = RealVarsModel()
        solver.solve(realvars.model)


def test_nolbvars():
    with pytest.raises(IncompatibleModelError):
        nolb = NoLbVarsModel()
        solver.solve(nolb.model)


def test_noubvars():
    with pytest.raises(IncompatibleModelError):
        noub = NoUbVarsModel()
        solver.solve(noub.model)


def test_quadcon():
    with pytest.raises(IncompatibleModelError):
        quadcon = QuadConModel()
        solver.solve(quadcon.model)


def test_nonlinearcon():
    with pytest.raises(IncompatibleModelError):
        nonlinearcon = NonlinearConModel()
        solver.solve(nonlinearcon.model)


def test_quadobj():
    with pytest.raises(IncompatibleModelError):
        quadobj = QuadObjModel()
        solver.solve(quadobj.model)


def test_nonlinearobj():
    with pytest.raises(IncompatibleModelError):
        nonlinearobj = NonlinearObjModel()
        solver.solve(nonlinearobj.model)


def test_infeasible_1():
    """
    raise_exception_on_nonoptimal_result = True (default)
    load_solutions = True (default)
    """
    with pytest.raises(NoOptimalSolutionError):
        infeasible = InfeasibleModel()
        solver.solve(infeasible.model)


def test_infeasible_2():
    """
    raise_exception_on_nonoptimal_result = False
    load_solutions = True (default)
    """
    with pytest.raises(NoFeasibleSolutionError):
        infeasible = InfeasibleModel()
        solver.solve(infeasible.model, raise_exception_on_nonoptimal_result=False)


def test_infeasible_3():
    """
    raise_exception_on_nonoptimal_result = False
    load_solutions = False
    """
    infeasible = InfeasibleModel()
    results = solver.solve(
        infeasible.model,
        raise_exception_on_nonoptimal_result=False,
        load_solutions=False,
    )

    assert (results.solution_status == SolutionStatus.infeasible) and (
        results.termination_condition == TerminationCondition.provenInfeasible
    )


def test_find_infeasible_subsystem(capfd):
    infeasible = InfeasibleModel()
    solver.solve(infeasible.model, find_infeasible_subsystem=True)
    captured = capfd.readouterr()
    assert '\ninfeasible_con\n' in captured.out


def test_inactive():
    """
    Solve a model with 2 constraints, 1 of which is inactive.
    Pyomo should only pass 1 constraint to CP-SAT.
    """
    inactive = InactiveConModel()
    solver.solve(inactive.model)
    assert len(solver._solver_model.proto.constraints) == 1


def test_constantobj():
    constantobj = ConstantObjModel()
    solver.solve(constantobj.model)
    assert pyo.value(constantobj.model.obj) == 500


def test_inactive_obj():
    with pytest.raises(ValueError):
        simple = SimpleModel()
        simple.model.obj.deactivate()
        solver.solve(simple.model)



import math
import pyomo.environ as pyo
from pyomo.contrib.solver.common.base import Availability
from pyomo.contrib.solver.common.results import SolutionStatus, TerminationCondition
import pyomo_cpsat
from model import SimpleModel

simple = SimpleModel()

solver = pyomo_cpsat.Cpsat()

results = solver.solve(
    simple.model,
    tee=True,
    threads=1,
    time_limit=100,
    rel_gap=0.0,
    abs_gap=1e-4,
    load_solutions=True,
)


## Start tests
def test_available():
    assert solver.available() == Availability.FullLicense


def test_version():
    assert len(solver.version()) == 3


def test_persistent():
    assert not solver.is_persistent()


def test_tee():
    assert solver._solver_solver.parameters.log_search_progress


def test_threads():
    assert solver._solver_solver.parameters.num_workers == 1


def test_time_limit():
    assert solver._solver_solver.parameters.max_time_in_seconds == 100


def test_rel_gap():
    assert math.isclose(solver._solver_solver.parameters.relative_gap_limit, 0.0)


def test_abs_gap():
    assert math.isclose(solver._solver_solver.parameters.absolute_gap_limit, 1e-4)


def test_solution_status():
    assert results.solution_status == SolutionStatus.optimal


def test_termination_condition():
    assert (
        results.termination_condition
        == TerminationCondition.convergenceCriteriaSatisfied
    )


def test_objective_value():
    assert pyo.value(simple.model.obj) == 196


def test_solution():
    assert (
        simple.model.x['chocolate'].value == 2
        and simple.model.x['vanilla'].value == 0
        and simple.model.x['matcha'].value == 8
    )



import math
import pyomo.environ as pyo
import pytest
from pyomo.contrib.solver.common.factory import SolverFactory
from pyomo.contrib.solver.common.base import Availability
from pyomo.contrib.solver.common.results import SolutionStatus, TerminationCondition
from pyomo.contrib.solver.common.util import (
    NoFeasibleSolutionError,
    NoOptimalSolutionError,
)
from pyomo.opt import Solution
from pytest import raises
import pyomo_cpsat
from model import SimpleModel

simple = SimpleModel()

# Fix x['vanilla'] to 2
simple.model.x['vanilla'].fix(2)

solver = pyomo_cpsat.Cpsat()

results = solver.solve(
    simple.model,
    tee=True,
    threads=1,
    time_limit=100,
    rel_gap=0.0,
    abs_gap=1e-4,
    load_solutions=True,
)


## Start tests
def test_solution_fix():
    assert (
        simple.model.x['chocolate'].value == 0
        and simple.model.x['vanilla'].value == 2
        and simple.model.x['matcha'].value == 7
    )


def test_objective_value_fix():
    assert pyo.value(simple.model.obj) == 193



"""
Compute optimal solution and optimal value for SimpleModel to use in tests
"""

import pyomo.environ as pyo
from model import SimpleModel

simple = SimpleModel()

#
# Solve model
#
solver = pyo.SolverFactory('glpk')
solver.solve(simple.model, tee=True)

print('\nObjective value:', pyo.value(simple.model.obj))

print('\nOptimal solution:')
for k in simple.model.K:
    print(f'x[{k}] = {simple.model.x[k].value}')

#
# Solve model with x['vanilla'] fixed to 1
#
simple.model.x['vanilla'].fix(2)
solver.solve(simple.model, tee=True)

print('\nObjective value:', pyo.value(simple.model.obj))

print('\nOptimal solution:')
for k in simple.model.K:
    print(f'x[{k}] = {simple.model.x[k].value}')
